import numpy as np


class CircularMemory(object):
    def __init__(self, memory_size=1000000):
        self.memory_size = memory_size
        self.memory = [None] * memory_size
        self.write_head = 0
        self.n_items = 0

    def __len__(self):
        return self.n_items

    def add(self, experience):
        position = self.write_head
        self.memory[position] = experience

        self.write_head = (self.write_head + 1) % self.memory_size
        self.n_items = min(self.n_items + 1, self.memory_size)

        return position

    def sample(self, sample_size=1):
        return np.random.choice(self.n_items, size=sample_size, replace=False)

    def __getitem__(self, item):
        if isinstance(item, list):
            return [self.memory[idx] for idx in item]
        return self.memory[item]



import numba
import numpy as np
import pandas as pd
from numba import prange
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.multiclass import OneVsRestClassifier
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils import check_X_y

from valiml.utils.numba_utils import create_progbar_numba


@numba.jit(nopython=True, parallel=True)
def optimize(X, y, alpha, max_iters, verbose):
    t = 0
    w = np.zeros((X.shape[1] + 1,), dtype=np.float32)

    if verbose == 1:
        bar = create_progbar_numba(max_iters)
    for n_iter in range(max_iters):
        for idx in range(X.shape[0]):
            t += 1
            lr = 1 / (t * alpha)
            if y[idx] * np.dot(w[1:], X[idx, :]) < 1:
                w[1:] -= lr * (alpha * w[1:] - y[idx] * X[idx])
            else:
                w[1:] -= lr * alpha * w[1:]

        if verbose == 1:
            bar.add(1)
    return w


def make_compute_kernel_matrix(kernel):
    @numba.jit(nopython=True, parallel=True)
    def _compute(X, kernel_matrix):
        for idx in range(X.shape[0]):
            kernel_matrix[idx:, idx] = kernel(X[idx:, :], X[idx, :])
            kernel_matrix[idx, idx:] = kernel_matrix[idx:, idx]

        return kernel_matrix
    return _compute


@numba.jit(nopython=True, parallel=True)
def optimize_kernel(X, y, alpha, max_iters, kernel_matrix, verbose):
    t = 0
    a = np.zeros((X.shape[0]), dtype=np.float32)

    if verbose == 1:
        bar = create_progbar_numba(max_iters)
    for n_iter in range(max_iters):
        for idx in range(X.shape[0]):
            t += 1
            lr = 1 / (t * alpha)

            value = lr * y[idx] * (a * y * kernel_matrix[idx, :]).sum()
            if value < 1:
                a[idx] = a[idx] + 1

        if verbose == 1:
            bar.add(1)
    return a


def make_rbf(gamma):
    @numba.jit(nopython=True, parallel=True)
    def rbf(X, w):
        result = np.zeros((X.shape[0],), dtype=np.float32)
        for idx in prange(X.shape[0]):
            result[idx] = -np.linalg.norm(X[idx, :] - w)**2 / (2 * gamma**2)
        return np.exp(result)
    return rbf


def make_apply_kernel(kernel):
    @numba.jit(nopython=True, parallel=True)
    def _kernel(X_train, X_test, a):
        result = np.zeros((X_test.shape[0],), dtype=np.float32)
        for idx in range(X_test.shape[0]):
            result[idx] = (kernel(X_train, X_test[idx]) * a).sum()
        return result
    return _kernel


class HingeSVM(BaseEstimator, ClassifierMixin):
    def __init__(self, optimizer='pegasos', max_iters=1000, alpha=0.1, kernel='linear', verbose=0, gamma=0.1):
        """

        :param optimizer: The optimizer to use. Choices include lm and gd.
        :param max_iters: Maximum number of iterations for convergence.
        :param alpha: The weight of the squared norm of w.
        :param n_inits: The number of initial starting position to try.
        :param regularizer: The regularizer of the respective optimizer.
                            Step size for gd and l2 regularization for lm.
        """
        self.optimizer = optimizer
        self.max_iters = max_iters
        self.alpha = alpha
        self.kernel = kernel
        self.verbose = verbose
        self.gamma = gamma

    def fit(self, X, y):
        X, y = check_X_y(X, y, multi_output=False, y_numeric=True, estimator='AdaBoost')

        self.coef_ = np.zeros((1, X.shape[1]))
        self.intercept_ = np.zeros((1, 1))

        X = X.astype('float32')
        y = 2 * y - 1
        if self.kernel == 'linear':
            w = optimize(X, y, self.alpha, self.max_iters, self.verbose)

            self.intercept_[0] = w[0]
            self.coef_[:] = w[1:]

            hinge_loss = np.clip(1 - self.decision_function(X) * y, 0, np.inf)
            self.support_vectors_ = X[hinge_loss > 0]
            self.dual_coef_ = hinge_loss[hinge_loss > 0]

        elif self.kernel == 'rbf':
            kernel = make_rbf(self.gamma)
            kernel_matrix = np.zeros((X.shape[0], X.shape[0]), dtype='float32')
            make_compute_kernel_matrix(kernel)(X, kernel_matrix)
            self.a = optimize_kernel(X, y, self.alpha, self.max_iters, kernel_matrix, self.verbose) * y
            self.X_train = X

        return self

    def decision_function(self, X):
        X = X.astype('float32')
        if self.kernel == 'linear':
            result = np.matmul(X, self.coef_[0])
        elif self.kernel == 'rbf':
            kernel = make_rbf(self.gamma)
            result = make_apply_kernel(kernel)(self.X_train, X, self.a)
        return result

    def predict(self, X):
        return np.sign(self.decision_function(X))


@numba.jit(nopython=True)
def test_progb():
    bar = create_progbar_numba(25)
    a = np.random.rand(2000, 10000)
    b = np.random.rand(10000, 2000)
    c = 0
    for idx in range(25):
        c += np.sum(a + b.T)

        bar.add(1)
    return c



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.metrics import accuracy_score
from sklearn.utils import check_array, check_X_y
from sklearn.model_selection import train_test_split


# TODO: Think of a way to use inheritance for multiple modes
from valiml.optimizers import lm
from valiml.utils import create_progbar, normalize


def solve_cubic(alpha):
    tmp = 3 * np.sqrt(3) * np.sqrt(27 * alpha**2 + 14 * alpha + 3)
    tmp = np.cbrt(tmp - 27 * alpha - 7)

    root = tmp / (3 * np.cbrt(2))
    root -= (2 * np.cbrt(2)) / (3 * tmp)
    root += 1/3

    return root


def make_prediction_sammer(estimator, X, n_labels):
    estimator_proba = estimator.predict_proba(X)
    np.clip(estimator_proba, np.finfo(estimator_proba.dtype).eps, None, out=estimator_proba)
    log_proba = np.log(estimator_proba)
    normalizer = log_proba.sum(axis=1)[:, np.newaxis]

    estimator_decision = log_proba - (1. / n_labels) * normalizer
    estimator_decision *= n_labels - 1

    estimator_prediction = estimator_decision.argmax(axis=1)

    return estimator_proba, estimator_decision, estimator_prediction


def one_hot_encode_samme(labels, n_labels):
    one_hot_encoded_labels = np.zeros((labels.shape[0], n_labels))
    one_hot_encoded_labels.fill(-1 / (n_labels - 1))

    for idx in range(labels.shape[0]):
        one_hot_encoded_labels[idx, labels[idx]] = 1
    return one_hot_encoded_labels


def get_train_and_validation(x_train, y_train, validation_split):
    if isinstance(validation_split, tuple):
        x_test, y_test = validation_split
    else:
        x_train, x_test, y_train, y_test = train_test_split(x_train, y_train,
                                                            test_size=validation_split,
                                                            shuffle=True)
    return x_train, x_test, y_train, y_test


class AdaBoost(BaseEstimator, ClassifierMixin):
    def __init__(self, base_estimator=None, n_estimators=200, algorithm='SAMME',
                 verbose=False, learning_rate=1.0, valid_boost=0, loss='exponential'):
        self.n_estimators = n_estimators
        self.verbose = verbose
        self.learning_rate = learning_rate
        self.algorithm = algorithm
        self.base_estimator = base_estimator
        self.valid_boost = valid_boost
        self.loss = loss
        self.initialize()

    def _get_alpha(self, classifier_decision, y, stump_decision, weighted_error):
        if self.algorithm == 'SAMME.R':
            return 1
        elif self.algorithm == 'SAMME':
            if self.loss == 'exponential':
                return np.log((1 - weighted_error) / weighted_error) + np.log(self.n_labels - 1)
            elif self.loss == 'logit':
                n_labels = self.n_labels

                def _logit(alpha, jacobian=True):
                    f = classifier_decision + alpha * stump_decision
                    exponential = np.exp(-(f * y).sum(axis=1) / n_labels)
                    loss = np.log(1 + exponential)

                    if jacobian:
                        derivative = -(1 - 1 / (1 + exponential)) * (y * stump_decision).sum(axis=1) / n_labels
                        return np.array([loss.mean()]), np.array([derivative.mean()])
                    return np.array([loss.mean()])

                alpha, loss = lm(np.array([0]), _logit, max_iters=1000, alpha=1e-5)
                return alpha[0]

    def initialize(self):
        self.validation_errors = None
        self.n_labels = -1
        self.losses = []
        self.errors = []
        self.weighted_errors = []
        self.estimators = []
        
    def _compute_loss(self, classifier_decision, y):
        if self.loss == 'exponential':
            return np.exp(-(classifier_decision * y).sum(axis=1) / self.n_labels).mean()
        elif self.loss == 'logit':
            return np.log(1 + np.exp(-(classifier_decision * y).sum(axis=1) / self.n_labels)).mean()

    def fit(self, X, y, validation_split=None):
        """
        :param x_train: features
        :param y_train: labels
        """

        x_train, y_train = check_X_y(X, y, multi_output=False, y_numeric=True, estimator='AdaBoost')

        check_array(x_train, accept_sparse=False, dtype='numeric', force_all_finite=True, ensure_2d=True,
                    estimator='AdaBoost')
        check_array(y_train, accept_sparse=False, dtype='numeric', force_all_finite=True, ensure_2d=False,
                    estimator='AdaBoost')

        self.initialize()

        if self.verbose:
            metrics = ['w_err', 'loss', 'err', 'acc']
            if validation_split is not None:
                metrics.append('val_acc')
            _bar = create_progbar(self.n_estimators, stateful_metrics=metrics)

        self.n_labels = len(np.unique(y_train))
        y_encoded = one_hot_encode_samme(y_train, self.n_labels)
        sample_weight = np.ones(x_train.shape[0]) / x_train.shape[0]
        classifier_decision = np.zeros((x_train.shape[0], self.n_labels))
        if validation_split is not None:
            x_train, x_test, y_train, y_test = get_train_and_validation(x_train, y_train, validation_split)
            self.validation_errors = []
            validation_decision = np.zeros((x_test.shape[0], self.n_labels))

        n_iter = 1
        retry_count = 0
        random_chance_error = (self.n_labels - 1) / self.n_labels
        used_samples = x_train.shape[0]

        if self.valid_boost:
            used_samples = int(x_train.shape[0] / 2)
            sample_weight[:used_samples] *= 2

        while n_iter <= self.n_estimators:
            if self.valid_boost and n_iter <= self.valid_boost * self.n_estimators:
                max_iter = self.n_estimators * self.valid_boost
                t = int(x_train.shape[0] / 2 + int((np.log(n_iter) / np.log(max_iter)) * x_train.shape[0] / 2))

                if t - used_samples > 0:
                    sample_weight[:used_samples] *= used_samples / t
                    sample_weight[used_samples:t] *= (t - used_samples) / t

                used_samples = t

            estimator = clone(self.base_estimator).fit(x_train[:used_samples], y_train[:used_samples],
                                                       sample_weight=sample_weight[:used_samples])

            if self.algorithm == 'SAMME':
                prediction_train = estimator.predict(x_train)
                decision_train = one_hot_encode_samme(prediction_train, self.n_labels)

                prediction_test = estimator.predict(x_test)
                decision_test = one_hot_encode_samme(prediction_test, self.n_labels)
            elif self.algorithm == 'SAMME.R':
                proba_train, decision_train, prediction_train = make_prediction_sammer(estimator,
                                                                                       x_train,
                                                                                       self.n_labels)
                _, decision_test, _ = make_prediction_sammer(estimator, x_test, self.n_labels)

            errors = prediction_train[:used_samples] != y_train[:used_samples]
            weighted_error = sample_weight[:used_samples][errors].sum()

            if weighted_error == 0:
                self.estimators.append((1.0, estimator))
                if self.verbose:
                    _bar.update(self.n_estimators, "|".join(metrics), [0, 0, 0, 1])
                break
            elif weighted_error >= random_chance_error:
                retry_count += 1
                if retry_count == 10:
                    break
                continue

            self.weighted_errors.append(weighted_error)

            alpha = self.learning_rate * self._get_alpha(classifier_decision[:used_samples],
                                                         y_encoded[:used_samples, :],
                                                         decision_train[:used_samples],
                                                         weighted_error)
            assert alpha >= 0

            self.estimators.append((alpha, estimator))

            classifier_decision += alpha * decision_train
            classifier_error = 1 - accuracy_score(y_train, classifier_decision.argmax(axis=1))
            self.errors.append(classifier_error)
            if validation_split is not None:
                validation_decision += alpha * decision_test
                validation_error = 1 - accuracy_score(y_test, validation_decision.argmax(axis=1))
                self.validation_errors.append(validation_error)

            if self.algorithm == 'SAMME':
                if self.loss == 'exponential':
                    sample_weight[:used_samples] *= np.exp(errors * alpha)
                elif self.loss == 'logit':
                    interim = (classifier_decision[:used_samples] * y_encoded[:used_samples]).sum(axis=1)
                    sample_weight[:used_samples] = (1 - 1 / (1 + np.exp(-interim / self.n_labels))) / self.n_labels
            elif self.algorithm == 'SAMME.R':
                if self.loss == 'exponential':
                    result = y_encoded[:used_samples] * np.log(proba_train[:used_samples])
                    result *= -(self.n_labels - 1) / self.n_labels
                    sample_weight[:used_samples] *= np.exp(result.sum(axis=1))
                else:
                    fy = (classifier_decision[:used_samples] * y_encoded).sum(axis=1)
                    sample_weight[:used_samples] = np.log(1 + np.exp(- (1 / self.n_labels) * fy))
            sample_weight[:used_samples] = normalize(sample_weight[:used_samples])

            self.losses.append(self._compute_loss(classifier_decision, y_encoded[:used_samples]))

            if self.verbose:
                values = {
                    'w_err': weighted_error,
                    'loss': self.losses[-1],
                    'err': classifier_error,
                    'acc': 1 - classifier_error
                }
                if validation_split is not None:
                    values['val_acc'] = 1 - validation_error

                _bar.update(
                    n_iter,
                    **values
                )

            n_iter += 1

        if self.verbose:
            _bar.update()
            _bar.finish(dirty=True)

        return self

    def decision_function(self, X):
        prediction = np.zeros((X.shape[0], self.n_labels))

        for alpha, estimator in self.estimators:
            if self.algorithm == 'SAMME':
                decision = one_hot_encode_samme(estimator.predict(X), self.n_labels)
                prediction += alpha * decision
            elif self.algorithm == 'SAMME.R':
                _, decision, _ = make_prediction_sammer(estimator, X, self.n_labels)
                prediction += alpha * decision

        return prediction

    def predict(self, X):
        decision = self.decision_function(X)
        return decision.argmax(axis=1)

    def predict_proba(self, X):
        if self.loss == 'exponential':
            prediction = self.decision_function(X)
            prediction /= self.n_labels
            prediction = np.exp((1. / (self.n_labels - 1)) * prediction)

            normalizer = prediction.sum(axis=1)[:, np.newaxis]
            normalizer[normalizer == 0.0] = 1.0
            prediction /= normalizer
            return prediction
        elif self.loss == 'logit':
            prediction = self.decision_function(X)
            prediction /= self.n_labels - 1
            prediction = np.exp(prediction) + 1
            normalizer = prediction.sum(axis=1)[:, np.newaxis]
            normalizer[normalizer == 0.0] = 1.0
            prediction /= normalizer
            return prediction



import numpy as np
from functools import wraps


def make_epsilon_greedy(agent, exploration_max=1.0, exploration_min=0.1, exploration_decay=0.995):
    agent.exploration_rate = exploration_max
    agent.exploration_min = exploration_min
    agent.exploration_decay = exploration_decay

    @wraps(agent.get_action)
    def get_action(self, state, available_actions=None):
        if np.random.rand() < self.exploration_rate:
            if available_actions is None:
                return self.available_actions.sample()
            return np.random.choice(list(available_actions))
        return agent.get_action(state, available_actions=available_actions)

    agent.get_action = get_action

    @wraps(agent.episode_end)
    def episode_end(self):
        agent.episode_end()
        self.exploration_rate *= self.exploration_decay
        self.exploration_rate = max(self.exploration_min, self.exploration_rate)

    return agent



from valiml.rl.agents.DQN import DQN
from valiml.rl.agents.DoubleDQN import DoubleDQN
from valiml.rl.agents.AdversarialDQN import AdversarialDQN
from valiml.rl.agents.GreedyEpsilonPolicy import make_epsilon_greedy

__all__ = [
    'DQN',
    'DoubleDQN',
    'AdversarialDQN',
    'make_epsilon_greedy'
]


import numpy as np

from valiml.rl.agents.DQN import DQN

class DoubleDQN(DQN):
    def __init__(self, training_model, inference_model, gamma=0.95, polyak_coef=0.01, shift_episodes=1):
        super(DoubleDQN, self).__init__(training_model, inference_model, gamma=gamma)
        self.target_model = inference_model.clone()
        self.polyak_coef = polyak_coef
        self.shift_episodes = shift_episodes

    def train(self, inputs, q_values, weights=None):
        loss = super(DoubleDQN, self).train(inputs, q_values, weights=weights)

        if self.n_training_steps == self.shift_episodes:
            agent_weights = self.inference_model.get_weights()
            target_weights = self.target_model.get_weights()

            new_weights = []
            for agent_layer_weights, target_layer_weights in zip(agent_weights, target_weights):
                new_layer_weights = self.polyak_coef * agent_layer_weights + \
                                    (1 - self.polyak_coef) * target_layer_weights
                new_weights.append(new_layer_weights)

            self.target_model.set_weights(new_weights)

        return loss

    def compute_q_target(self, state, action, reward, next_state, terminal):
        target_action = np.argmax(self.target_model.predict(next_state)[0])
        return reward if terminal else reward + self.gamma * self.inference_model.predict(next_state)[0][target_action]



import numpy as np

from valiml.rl.agents.base import QLearningAgent

try:
    from keras.models import clone_model
except ImportError as e:
    from copy import deepcopy as clone_model


def restricted_argmax(array, limited_indexes):
    limited_indexes = list(limited_indexes)
    best_idx = limited_indexes[0]
    if len(limited_indexes) > 1:
        for idx in limited_indexes[1:]:
            if array[0][idx] > array[0][best_idx]:
                best_idx = idx

    return best_idx


class DQN(QLearningAgent):
    def __init__(self, training_model, inference_model, gamma=0.95):
        self.training_model = training_model
        self.inference_model = inference_model
        self.gamma = gamma
        self.n_training_steps = 0

    def get_action(self, state, available_actions=None):
        prediction = self.get_q_values(state)

        if available_actions is None:
            return np.argmax(prediction)
        return restricted_argmax(prediction, available_actions)

    def train(self, inputs, q_values, weights=None):
        if weights is None:
            weights = np.ones((inputs[0].shape[0], 1))

        history = self.training_model.fit([*inputs, weights], q_values, verbose=0)

        self.n_training_steps += 1

        return history.history['loss'][-1]

    def compute_q_target(self, state, action, reward, next_state, terminal):
        return reward if terminal else reward + self.gamma * np.amax(self.inference_model.predict(next_state)[0])

    def get_q_values(self, states):
        return self.inference_model.predict(states)

    def clone(self):
        return clone_model(self.inference_model)



import numpy as np
from valiml.rl.agents.DoubleDQN import DoubleDQN


class AdversarialDQN(DoubleDQN):
    def __init__(self, training_model, inference_model, **kwargs):
        super(AdversarialDQN, self).__init__(training_model, inference_model, **kwargs)
        self.other_player = None

    def set_other_player(self, agent):
        self.other_player = agent

    def compute_q_target(self, state, action, reward, next_state, terminal):
        target_action = np.argmax(self.other_player.target_model.predict(next_state)[0])
        opponent_reward =  self.other_player.inference_model.predict(next_state)[0][target_action]
        return reward if terminal else reward - self.gamma * opponent_reward



import numpy as np


class PrioritizedExperienceReplay(object):
    def __init__(self, beta=0.5):
        self.beta = beta

    def train(self, agent, prioritized_memory, batch_size=20):
        if len(prioritized_memory) < batch_size:
            return

        rewards = []
        inputs = []
        outputs = []
        weights = []
        for idx in prioritized_memory.sample(sample_size=batch_size):
            state, action, reward, state_next, terminal = prioritized_memory[idx]
            q_update = agent.compute_q_target(state, action, reward, state_next, terminal)
            q_values = agent.get_q_values(state)
            q_diff = q_update - q_values[0][action]
            q_values[0][action] = q_update

            prioritized_memory.priority_update(idx, np.abs(q_diff))

            priority = prioritized_memory.get_priority(idx)
            w = np.power((1 / len(prioritized_memory)) * (1 / priority), self.beta)

            rewards.append(reward)

            inputs.append(state)
            outputs.append(q_values)
            weights.append(w)

        loss = agent.train(
            [np.concatenate(inputs, axis=0)],
            np.concatenate(outputs, axis=0),
            weights=np.array(weights) / np.sum(weights)
        )

        return loss, np.mean(rewards)



from progressbar import ProgressBar, UnknownLength
from progressbar import SimpleProgress, DynamicMessage, Bar, AdaptiveETA


def marker(progress, data, width):
    if progress.max_value is not UnknownLength and progress.max_value > 0:
        length = int(progress.value / progress.max_value * width)
        return "".join((['='] * max(0, length - 1) + ['>']))
    else:
        return '='


class ETAFormat:
    def format(self, **data):
        eta_seconds = data['eta_seconds']
        if eta_seconds > 3600:
            eta_format = ('%d:%02d:%02d' %
                          (eta_seconds // 3600, (eta_seconds % 3600) // 60, eta_seconds % 60))
        elif eta_seconds > 60:
            eta_format = '%d:%02d' % (eta_seconds // 60, eta_seconds % 60)
        else:
            eta_format = '%ds' % eta_seconds

        return ' - ETA: %s' % eta_format


class FinishedETAFormat:
    def format(self, **data):
        time_per_unit = data['seconds_elapsed'] / data['value']
        if time_per_unit >= 1:
            return ' %.0fs/step' % time_per_unit
        elif time_per_unit >= 1e-3:
            return ' %.0fms/step' % (time_per_unit * 1e3)
        else:
            return ' %.0fus/step' % (time_per_unit * 1e6)


class ValueFormat:
    def format(self, **data):
        if data['value'] is None:
            return '-'
        return ' - {name} {value:.3f}'.format(**data)


def create_progbar_keras(max_value, stateful_metrics=[]):
    widgets = [
        SimpleProgress(format='{value}/{max_value}', new_style=True), ' ',
        Bar(marker=marker, left='[', right=']', fill='.'),
        AdaptiveETA(
            format_not_started='- ETA: --',
            format_finished=FinishedETAFormat(),
            format=ETAFormat(),
            format_zero=' - ETA: 0s',
            format_NA='- ETA: N/A',
            new_style=True
        )
    ]

    for metric in stateful_metrics:
        widgets.append(DynamicMessage(name=metric, format=ValueFormat()))
    return ProgressBar(max_value=max_value, widgets=widgets, redirect_stdout=True)



import numba
import numpy as np
from numba import prange
from sklearn.base import BaseEstimator, ClassifierMixin

from valiml.utils.numba_utils import entropy


@numba.jit(nopython=True, cache=True)
def normalize(v):
    return v / v.sum()


@numba.jit(nopython=True, parallel=True, cache=True)
def fit_stump(X, Y, feature_ids, n_labels, reg_entropy=0, sample_weight=None):
    n_features = len(feature_ids)

    weighted_errors = np.zeros((n_features,), dtype=np.float64)
    split_values = np.zeros((n_features,), dtype=np.float64)
    left_probabilities = np.zeros((n_features, n_labels), dtype=np.float64)
    right_probabilities = np.zeros((n_features, n_labels), dtype=np.float64)

    for fidx in prange(n_features):
        feature_idx = feature_ids[fidx]

        sorted_data_mask = X[:, feature_idx].argsort()
        x = X[:, feature_idx][sorted_data_mask]
        y = Y[sorted_data_mask]
        sorted_sample_weight = sample_weight[sorted_data_mask]

        sample_sum_left = np.zeros(n_labels)
        sample_sum_right = np.zeros(n_labels)
        for idx in range(0, y.shape[0]):
            sample_sum_right[y[idx]] += sorted_sample_weight[idx]

        total_weighted_sum = sample_sum_right.sum()
        if reg_entropy == 0:
            prob_left = normalize(np.random.random((n_labels,)))
        else:
            prob_left = np.zeros((n_labels,), dtype=np.float64)
            prob_left[np.random.choice(n_labels, size=1)] = 1.

        prob_right = normalize(sample_sum_right)
        weighted_error = total_weighted_sum - sample_sum_right.max()
        weighted_error += entropy(prob_right)

        best_stump = (weighted_error, prob_left, prob_right, x[0])

        for idx in range(1, y.shape[0]):
            sample_sum_left[y[idx]] += sorted_sample_weight[idx]
            sample_sum_right[y[idx]] -= sorted_sample_weight[idx]

            if x[idx - 1] != x[idx]:
                prob_left = normalize(sample_sum_left)
                prob_right = normalize(sample_sum_right)

                left_label = prob_left.argmax()
                right_label = prob_right.argmax()

                weighted_error = total_weighted_sum - sample_sum_left[left_label] - sample_sum_right[right_label]

                weighted_error += reg_entropy * (
                    entropy(prob_left) * sample_sum_left.sum() / total_weighted_sum +
                    entropy(prob_right) * sample_sum_right.sum() / total_weighted_sum
                )

                if weighted_error < best_stump[0]:
                    best_split_value = x[idx - 1] + (x[idx] - x[idx - 1]) / 2
                    best_stump = (weighted_error, prob_left, prob_right, best_split_value)

        weighted_errors[fidx], left_probabilities[fidx], right_probabilities[fidx], split_values[fidx] = best_stump

    best_idx = weighted_errors.argmin()
    return (weighted_errors[best_idx], feature_ids[best_idx],
            split_values[best_idx],
            left_probabilities[best_idx], right_probabilities[best_idx])


class DecisionStump(BaseEstimator, ClassifierMixin):
    def __init__(self, n_random_features=None, reg_entropy=0):
        self.feature_idx = -1
        self.left_probabilities = None
        self.right_probabilities = None
        self.split_value = 0
        self.n_labels = 0
        self.n_random_features = n_random_features
        self.classes_ = 0
        self.reg_entropy = reg_entropy
        self.left_label = 0
        self.right_label = 0

    def fit(self, X, y, sample_weight=None):
        self.classes_ = np.unique(y)
        self.n_labels = len(self.classes_)

        if self.n_random_features is not None:
            n_random_features = min(self.n_random_features, X.shape[1])
            features_subset = np.random.choice(X.shape[1], n_random_features, replace=False)
        else:
            features_subset = list(range(X.shape[1]))

        if sample_weight is None:
            sample_weight = np.zeros(X.shape[0])
            sample_weight.fill(1 / X.shape[0])

        error, self.feature_idx, self.split_value, self.left_probabilities, self.right_probabilities = fit_stump(
            X, y, features_subset, self.n_labels, reg_entropy=self.reg_entropy, sample_weight=sample_weight
        )

        self.left_label = self.left_probabilities.argmax()
        self.right_label = self.right_probabilities.argmax()

        return self

    def predict_proba(self, X):
        probability = np.zeros((X.shape[0], self.n_labels))
        mask = X[:, self.feature_idx] < self.split_value

        probability[mask, :] = self.left_probabilities
        probability[~mask, :] = self.right_probabilities

        np.clip(probability, np.finfo(probability.dtype).eps, None, out=probability)

        return probability

    def predict(self, X):
        prediction = np.zeros(X.shape[0], dtype=int)
        mask = X[:, self.feature_idx] < self.split_value

        prediction[mask] = self.left_label
        prediction[~mask] = self.right_label
        return prediction

    def __lt__(self, other):
        return self.split_value < other.split_value



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os.path import join

try:
    long_description = open("README.rst").read()
except IOError:
    long_description = ""

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="valiml",
    version="0.1.17",
    description="Extension of sklearn with tweaked implementation of common machine learning algorithms for self-use.",
    license="MIT",
    author="Valentin-Bogdan Rosca",
    packages=find_packages(),
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
    ],
    setup_requires=requirements,
    install_requires=["cffi"],
    cffi_modules=[join('valiml', 'utils', 'src', 'cffi.py:ffibuilder')],
    zip_safe=True
)



from abc import abstractmethod


class Agent(object):
    @abstractmethod
    def get_action(self, state, available_actions=None):
        raise NotImplemented()


class LearningAgent(Agent):
    @abstractmethod
    def episode_end(self):
        raise NotImplemented()

    @abstractmethod
    def train(self, inputs, outputs, weights=None):
        raise NotImplemented()


class QLearningAgent(Agent):
    @abstractmethod
    def train(self, inputs, q_values, weights=None):
        raise NotImplemented()

    @abstractmethod
    def compute_q_target(self, state, action, reward, next_state, terminal):
        raise NotImplemented()



import sys

from valiml.utils.progbar import create_progbar_keras


def normalize(x):
    return x / x.sum()


def create_progbar(target, stateful_metrics=None, numba=False):
    if numba:
        from valiml.utils.numba_utils import ProgressBar

        dynamic_display = ((hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()) or 'ipykernel' in sys.modules)
        stateful_metrics = '|'.join(set(stateful_metrics)) if stateful_metrics else ''

        return ProgressBar(target,
                           width=30,
                           verbose=1,
                           stateful_metrics=stateful_metrics,
                           dynamic_display=dynamic_display
                           )
    else:
        return create_progbar_keras(target, stateful_metrics=stateful_metrics)



import numpy as np

from valiml.rl.memory import CircularMemory
from valiml.rl.utils import SumSegmentTree


class PrioritizedMemory(CircularMemory):
    def __init__(self, memory_size=100000, alpha=1.0, eps=1e-7):
        super(PrioritizedMemory, self).__init__(memory_size=memory_size)
        self.sum_segment_tree = SumSegmentTree(memory_size)
        self.max_priority = 1
        self.eps = eps
        self.alpha = alpha

    def __len__(self):
        return self.n_items

    def add(self, experience):
        position = super(PrioritizedMemory, self).add(experience)
        priority = (self.max_priority + self.eps) ** self.alpha
        self.sum_segment_tree.update(position, priority)

    def priority_update(self, index, priority):
        self.max_priority = max(self.max_priority, priority)
        priority = (priority + self.eps) ** self.alpha
        self.sum_segment_tree.update(index, priority)

    def get_priority(self, index):
        return self.sum_segment_tree.get(index)

    def sample(self, sample_size=1, use_priority=True):
        if use_priority:
            return [self.sum_segment_tree.lower_bound(np.random.random()) for _ in range(sample_size)]
        return super(PrioritizedMemory, self).sample(sample_size=sample_size)



import unittest


class SumSegmentTree(object):
    def __init__(self, size):
        self.size = size
        self.limit = 1
        while self.limit < size:
            self.limit *= 2

        self.tree = [0] * (2 * self.limit)

    def _update(self, index, start, end, position, value):
        if start == end:
            old_value = self.tree[index]
            self.tree[index] = value
            return old_value

        self.tree[index] += value

        middle = int((start + end) // 2)

        if position <= middle:
            old_value = self._update(2 * index, start, middle, position, value)
        else:
            old_value = self._update(2 * index + 1, middle + 1, end, position, value)

        self.tree[index] -= old_value
        return old_value

    def update(self, index, value):
        self._update(1, 1, self.limit, index + 1, value)

    def _get(self, index, start, end, position):
        if start == end:
            return self.tree[index]

        middle = int((start + end) // 2)
        if position <= middle:
            return self._get(2 * index, start, middle, position)
        else:
            return self._get(2 * index + 1, middle + 1, end, position)

    def get(self, index):
        return self._get(1, 1, self.limit, index + 1)

    def _lower_bound(self, index, start, end, cumulative_sum):
        if start == end:
            return start

        left_priority = self.tree[2 * index] / self.tree[1]
        middle = (start + end) // 2

        if cumulative_sum <= left_priority:
            return self._lower_bound(2 * index, start, middle, cumulative_sum)
        else:
            return self._lower_bound(2 * index + 1, middle + 1, end, cumulative_sum - left_priority)

    def lower_bound(self, cumulative_sum):
        return self._lower_bound(1, 1, self.limit, cumulative_sum) - 1


class SumSegmentTreeTests(unittest.TestCase):
    def test_update(self):
        tree = SumSegmentTree(1000)
        tree.update(1, 100)
        tree.update(2, 10)
        self.assertEqual(tree.get(1), 100)

    def test_lower_bound(self):
        tree = SumSegmentTree(1000)

        multiplier = 2.5
        tree.update(0, 0.2 * multiplier)
        tree.update(1, 0.5 * multiplier)
        tree.update(2, 0.1 * multiplier)
        tree.update(3, 0.05 * multiplier)
        tree.update(4, 0.05 * multiplier)
        tree.update(5, 0.1 * multiplier)

        self.assertEqual(tree.lower_bound(0.2), 0)
        self.assertEqual(tree.lower_bound(0.3), 1)
        self.assertEqual(tree.lower_bound(0.7), 1)
        self.assertEqual(tree.lower_bound(0.71), 2)


if __name__ == '__main__':
    unittest.main()
